// Version using 2x Si5351 outputs, EEPROM function and 25MHz crystal

// Si5351a outputs:
// CLK0 not used 
// CLK1 ( for 1.IF tuning 69.9-70.6 MHz ) Lo = 59.2-59.9 MHz 
// CLK2 ( fix 2.IF ) = 10.7 MHz

// port #2 and #3 for tuning by increment generator 
// push #8 for storing frequency in EEPROM 
// push #9 for tuning step selection 
// I2C-Bus #A4 (SDA) and #A5 (SCL) for Si5351 and OLED 

/**************************************************************************
  Using 10kHz to 225MHz VFO / RF Generator with Si5351 and Arduino Nano 
  originated by J. CesarSound - ver 1.0 - Dec/2020 
  Program mofified and extended by Klaus Hirschelmann, DJ7OO 
***************************************************************************/

//Libraries
#include <Wire.h>                 //IDE Standard
#include <Rotary.h>               //Ben Buxton https://github.com/brianlow/Rotary
#include <si5351.h>               //Etherkit https://github.com/etherkit/Si5351Arduino
#include <Adafruit_GFX.h>         //Adafruit GFX https://github.com/adafruit/Adafruit-GFX-Library
#include <Adafruit_SSD1306.h>     //Adafruit SSD1306 https://github.com/adafruit/Adafruit_SSD1306
#include "EEPROM.h" 

//User preferences
//------------------------------------------------------------------------------------------------------------
#define IF  -10700           //Enter your IF frequency in KHz 
#define FREQ_INIT  70150000  //Enter your initial frequency at startup in Hz.
#define XT_CAL_F   58000     //Si5351 calibration factor, adjust to get exatcly 10MHz. Increasing this value will decreases the frequency and vice versa.
#define tunestep   9         //Change the pin used by encoder push button if you want.
#define eestore    8         //store current frequency in EEPROM 

Rotary r = Rotary(2, 3); 
Adafruit_SSD1306 display = Adafruit_SSD1306(128, 64, &Wire);
Si5351 si5351;

unsigned long freq = FREQ_INIT;
unsigned long freqold, fstep;
long interfreq = IF;
long cal = XT_CAL_F;
byte encoder = 1;
byte stp;
unsigned int period = 100;   //millis display active
unsigned long time_now = 0;  //millis display active
int press;

ISR(PCINT2_vect) {
  char result = r.process();
  if (result == DIR_CW) set_frequency(1);
  else if (result == DIR_CCW) set_frequency(-1);
}

//////////////////////////////////////////////////////////////
void EEPROMWritelong(int address,long value)
      {
      //Decomposition from a long to 4 bytes by using bitshift.
      //One = Most significant -> Four = Least significant byte
      byte four = (value & 0xFF);
      byte three = ((value >> 8) & 0xFF);
      byte two = ((value >> 16) & 0xFF);
      byte one = ((value >> 24) & 0xFF);

      //Write the 4 bytes into the eeprom memory.
      EEPROM.write(address, four);
      EEPROM.write(address + 1, three);
      EEPROM.write(address + 2, two);
      EEPROM.write(address + 3, one);
      }

//////////////////////////////////////////////////////////////
//This function will return a 4 byte (32bit) long from the eeprom
//at the specified address to address + 3.
long EEPROMReadlong(long address)
      {
      //Read the 4 bytes from the eeprom memory.
      long four = EEPROM.read(address);
      long three = EEPROM.read(address + 1);
      long two = EEPROM.read(address + 2);
      long one = EEPROM.read(address + 3);

      //Return the recomposed long by using bitshift.
      return ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);
      }

/////////////////////////////////////////////////////
void set_frequency(short dir) {
  if (encoder == 1) {                         //Up/Down frequency
    if (dir == 1) freq = freq + fstep;
    if (freq >= 70600000) freq = 70600000;
    if (dir == -1) freq = freq - fstep;
    if (fstep == 1000000 && freq <= 1000000) freq = 1000000;
    else if (freq < 69900000) freq = 69900000;
  }
}

//////////////////////////////////////////////////////////////
void setup() {
  Wire.begin();
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setTextColor(WHITE);
  display.display();
  startup_text();  
  pinMode(2, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  pinMode(tunestep, INPUT_PULLUP);
  pinMode(eestore, INPUT_PULLUP);  

  freq = EEPROMReadlong(0); 
  if(freq<69900000||freq>70600000) 
  { freq=70150000; } // use if value from EEPROM is invalid  
 
  si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  si5351.set_correction(cal, SI5351_PLL_INPUT_XO);
  si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_2MA);
  si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_2MA);  
  si5351.output_enable(SI5351_CLK0, 0);       //1 - Enable / 0 - Disable CLK
  si5351.output_enable(SI5351_CLK1, 1);
  si5351.output_enable(SI5351_CLK2, 1);

  PCICR |= (1 << PCIE2);
  PCMSK2 |= (1 << PCINT18) | (1 << PCINT19);
  sei();

  stp = 3;
  setstep();
  layout();
  displayfreq();
}

/////////////////////////////////////////////////////////
void loop() {
  press = digitalRead(eestore);
  if (press == LOW)
  {
  long address=0;
  EEPROMWritelong(address, freq);
  address+=4;
  } 
   
  if (freqold != freq) {
    time_now = millis();
    tunegen();
    freqold = freq;
  }

  if (digitalRead(tunestep) == LOW) {
    time_now = (millis() + 300);
    setstep();
    delay(300);
  }

  if ((time_now + period) > millis()) {
    displayfreq();
    layout();
  }
}

////////////////////////////////////////////////////////////////////
void tunegen() {
  si5351.set_freq((freq + (interfreq * 1000ULL)) * 100ULL, SI5351_CLK1);
  si5351.set_freq(10700000 * 100ULL, SI5351_CLK2);
}

////////////////////////////////////////////////////////////////////
void displayfreq() {
  display.clearDisplay();
  display.setTextSize(2);
  char buffer[15] = "";
  display.print(buffer);
}

///////////////////////////////////////////////////////////////////////
void setstep() {
  switch (stp) {
    case 1:
      stp = 2;
      fstep = 10;
      break;
    case 2:
      stp = 3;
      fstep = 100;
      break;
    case 3:
      stp = 4;
      fstep = 1000;
      break;
    case 4:
      stp = 5;
      fstep = 5000;
      break;
    case 5:
      stp = 6;
      fstep = 10000;
      break;
    case 6:
      stp = 1;
      fstep = 100000;
      break;
   }
}

//////////////////////////////////////////////////////////////////
void layout() {
  display.setTextColor(WHITE);
  display.drawLine(0, 20, 127, 20, WHITE);
  display.drawLine(0, 43, 127, 43, WHITE);
  display.setTextSize(2);
  display.setCursor(2, 50);
  display.print("STEP:");
  if (stp == 2) display.print("10"); 
  if (stp == 3) display.print("100"); 
  if (stp == 4) display.print("1");
  if (stp == 5) display.print("5"); 
  if (stp == 6) display.print("10"); 
  if (stp == 1) display.print("100");
  display.setTextSize(1);
  if(stp==2|stp==3) display.print(" Hz"); 
  else display.print(" KHz");
  display.setTextSize(2);
  display.setCursor(0,0);
  display.print(" 4m-RX");
  display.setCursor(20,25);
  //display.print("10");
  float freq2=(freq/100); 
  freq2=(freq2/10000);
  display.print(freq2,4);
  display.display();
}

/////////////////////////////////////////////////////////
void startup_text() {
  display.setTextSize(1);
  display.setCursor(4, 5);
  display.print(">> 4m-RX <<");
  display.setCursor(4, 20);
  display.print(">> by DJ7OO <<");
  display.setCursor(4, 35);
  display.print(" based on code ");
  display.setCursor(4, 50);
  display.print(" by J.CesarSound ");
  display.display();
  delay(3000);
  display.clearDisplay();
}
